@@ -4,8 +4,7 @@
 -- RSA standard and RFC 3447.
 module Codec.Crypto.RSA(
        -- * Keys and key generations
-         PublicKey(..), PrivateKey(..)
-       , generateKeyPair
+       generateKeyPair
        -- * High-level encryption and signing functions
        , encrypt
        , decrypt
@@ -53,7 +52,9 @@ import qualified Data.ByteString.Lazy as BS
 import Data.Digest.Pure.SHA
 import Data.Int
 import Data.Word
-import System.Random
+import Crypto.Random
+import Crypto.Types.PubKey.RSA
+import Control.Monad.CryptoRandom
 
 #ifdef USE_BINARY
 import Data.Binary
@@ -65,27 +66,13 @@ import Data.Binary.Get
 import Data.Digest.Pure.MD5
 #endif
 
-data PublicKey = PublicKey { 
-    public_size :: Int64   -- ^The size of the RSA modulus, in bytes.
-  , public_n    :: Integer -- ^The RSA modulus.
-  , public_e    :: Integer -- ^The public exponent.
-  }
- deriving (Show)
-
-data PrivateKey = PrivateKey {
-    private_size :: Int64   -- ^The size of the RSA modulus, in bytes.
-  , private_n    :: Integer -- ^The RSA modulus.
-  , private_d    :: Integer -- ^The private exponent.
-  }
- deriving (Show)
-
 #ifdef USE_BINARY
 instance Binary PublicKey where
   put pk = do putLazyByteString $ i2osp (public_size pk) 8
               putLazyByteString $ i2osp (public_n pk)    (public_size pk)
   get    = do len <- (fromIntegral . os2ip) `fmap` getLazyByteString 8
               n   <- os2ip `fmap` getLazyByteString len
-              return $ PublicKey len n 65537 
+              return $ PublicKey (fromIntegral len) n 65537
 
 instance Binary PrivateKey where
   put pk = do putLazyByteString $ i2osp (private_size pk) 8
@@ -94,7 +81,14 @@ instance Binary PrivateKey where
   get    = do len <- (fromIntegral . os2ip) `fmap` getLazyByteString 8
               n   <- os2ip `fmap` getLazyByteString len
               d   <- os2ip `fmap` getLazyByteString len
-              return $ PrivateKey len n d            
+              return $ PrivateKey { private_size = fromIntegral len
+                                  , private_n    = n
+                                  , private_d    = d
+                                  , private_p    = 0
+                                  , private_q    = 0
+                                  , private_qinv = 0
+                                  , private_dP   = 0
+                                  , private_dQ   = 0 }
 #endif
 
 type HashFunction = ByteString -> ByteString
@@ -119,9 +113,9 @@ type MGF          = ByteString -> Int64 -> ByteString
 -- |Randomly generate a key pair of the given modulus length (in bits) to
 -- use in any of the following functions. Use of a good random number 
 -- generator is of considerable importance when using this function; the 
--- input RandomGen should never be used again for any other purpose.
-generateKeyPair :: RandomGen g => g -> Int -> (PublicKey, PrivateKey, g)
-generateKeyPair g sizeBits = (PublicKey kLen n e, PrivateKey kLen n d, g')
+-- input CryptoRandomGen should never be used again for any other purpose.
+generateKeyPair :: CryptoRandomGen g => g -> Int -> (PublicKey, PrivateKey, g)
+generateKeyPair g sizeBits = (PublicKey kLen n e, privateKey, g')
  where
   kLen       = fromIntegral $ sizeBits `div` 8
   (p, q, g') = generate_pq g kLen
@@ -129,6 +123,14 @@ generateKeyPair g sizeBits = (PublicKey kLen n e, PrivateKey kLen n d, g')
   phi        = (p - 1) * (q - 1)
   e          = 65537
   d          = modular_inverse e phi 
+  privateKey = PrivateKey { private_size = kLen
+                          , private_n    = n
+                          , private_d    = d
+                          , private_p    = 0
+                          , private_q    = 0
+                          , private_qinv = 0
+                          , private_dP   = 0
+                          , private_dQ   = 0 }
 
 data EncryptionOptions = 
     UseOAEP {
@@ -151,7 +153,7 @@ instance Show EncryptionOptions where
 -- function, and not adding a label). If the message is longer than the 
 -- underlying encryption function can support, it is broken up into parts
 -- and each part is encrypted.
-encrypt :: RandomGen g => g -> PublicKey -> ByteString -> (ByteString, g)
+encrypt :: CryptoRandomGen g => g -> PublicKey -> ByteString -> (ByteString, g)
 encrypt = encrypt' (UseOAEP sha256' (generate_MGF1 sha256') BS.empty)
 
 -- |Decrypt an arbitrarily-sized message using the defaults for RSA
@@ -182,21 +184,20 @@ verify :: PublicKey -> ByteString -> ByteString -> Bool
 verify = rsassa_pkcs1_v1_5_verify ha_SHA256
 
 -- |Encrypt an arbitrarily-sized message using the given options.
-encrypt' :: RandomGen g => 
+encrypt' :: CryptoRandomGen g => 
             EncryptionOptions -> g -> PublicKey -> ByteString -> 
             (ByteString, g)
 encrypt' (UseOAEP hash mgf l) gen pub m = foldl enc1 (BS.empty, gen) chunks
  where
   hLen              = BS.length $ hash BS.empty
-  chunkSize         = public_size pub - (2 * hLen) - 2
+  chunkSize         = (fromIntegral $ public_size pub) - (2 * hLen) - 2
   chunks            = chunkify chunkSize m
-  enc1 (!res, !g) !cur = let (seed, g') = random g
-                             !newc = rsaes_oaep_encrypt hash mgf pub seed l cur
+  enc1 (!res, !g) !cur = let !(!newc,!g') = rsaes_oaep_encrypt g hash mgf pub l cur
                          in (res `BS.append` newc, g')
 encrypt' UsePKCS1_v1_5        gen pub m = foldl enc1 (BS.empty, gen) chunks
  where
   chunkSize         = public_size pub - 11
-  chunks            = chunkify chunkSize m
+  chunks            = chunkify (fromIntegral chunkSize) m
   enc1 (!res, !g) !cur = let (!newc, g')=rsaes_pkcs1_v1_5_encrypt g pub cur
                          in (res `BS.append` newc, g')
 
@@ -206,7 +207,7 @@ encrypt' UsePKCS1_v1_5        gen pub m = foldl enc1 (BS.empty, gen) chunks
 decrypt' :: EncryptionOptions -> PrivateKey -> ByteString -> ByteString
 decrypt' opts priv cipher = BS.concat $ map decryptor chunks
  where
-  chunks = chunkify (private_size priv) cipher
+  chunks = chunkify (fromIntegral $ private_size priv) cipher
   decryptor = case opts of
                 UseOAEP hash mgf l -> rsaes_oaep_decrypt hash mgf priv l
                 UsePKCS1_v1_5      -> rsaes_pkcs1_v1_5_decrypt priv
@@ -234,17 +235,18 @@ decrypt' opts priv cipher = BS.concat $ map decryptor chunks
 -- I have not put in a check for the length of the label, because I don't
 -- expect you to use more than 2^32 bytes. So don't make me regret that, eh?
 --
-rsaes_oaep_encrypt :: HashFunction -> MGF -> 
-                      PublicKey -> Integer -> ByteString -> ByteString ->
-                      ByteString
-rsaes_oaep_encrypt hash mgf k seed_int l m
+rsaes_oaep_encrypt :: CryptoRandomGen g => g -> HashFunction -> MGF -> 
+                      PublicKey -> ByteString -> ByteString ->
+                      (ByteString,g)
+rsaes_oaep_encrypt g hash mgf k l m
   | message_too_long = error "message too long (rsaes_oaep_encrypt)"
-  | otherwise        = c
+  | otherwise        = (c,g')
  where
   mLen = BS.length m -- Int64
   hLen = BS.length $ hash BS.empty -- Int64
-  kLen = public_size k
-  seed = i2osp seed_int hLen
+  kLen = fromIntegral $ public_size k
+  (seedStrict,g') = throwLeft $ genBytes (fromIntegral hLen) g
+  seed = BS.fromChunks [seedStrict]
   -- Step 1
   message_too_long = mLen > (kLen - (2 * hLen) - 2)
   -- Step 2
@@ -259,7 +261,7 @@ rsaes_oaep_encrypt hash mgf k seed_int l m
   -- Step 3
   m_ip       = os2ip em
   c_ip       = rsa_ep (public_n k) (public_e k) m_ip
-  c          = i2osp c_ip kLen
+  c          = i2osp c_ip (fromIntegral kLen)
 
 -- |The generalized implementation of RSAES-OAEP-DECRYPT. Again, 'decrypt'
 -- initializes this with a pretty good set of defaults if you don't understand
@@ -294,8 +296,8 @@ rsaes_oaep_decrypt hash mgf k l c
   hLen = BS.length $ hash BS.empty
   kLen = private_size k
   -- Step 1
-  bad_message_len = BS.length c /= kLen
-  bad_hash_len    = kLen < ((2 * hLen) + 2)
+  bad_message_len = BS.length c /= fromIntegral kLen
+  bad_hash_len    = fromIntegral kLen < ((2 * hLen) + 2)
   -- Step 2
   c_ip            = os2ip c
   m_ip            = rsa_dp (private_n k) (private_d k) c_ip
@@ -306,7 +308,7 @@ rsaes_oaep_decrypt hash mgf k l c
   (maskedSeed, maskedDB) = BS.splitAt hLen msandmdb
   seedMask               = mgf maskedDB hLen
   seed                   = maskedSeed `xorBS` seedMask
-  dbMask                 = mgf seed (kLen - hLen - 1)
+  dbMask                 = mgf seed (fromIntegral kLen - hLen - 1)
   db                     = maskedDB `xorBS` dbMask
   (lHash', ps1m)         = BS.splitAt hLen db
   one_m                  = BS.dropWhile (== 0) ps1m
@@ -332,14 +334,14 @@ rsaes_oaep_decrypt hash mgf k l c
 -- randomness, and (b) choose a decent instance of RandomGen for passing to
 -- this function.
 --
-rsaes_pkcs1_v1_5_encrypt :: RandomGen g => 
+rsaes_pkcs1_v1_5_encrypt :: CryptoRandomGen g => 
                             g -> PublicKey -> ByteString -> 
                             (ByteString, g)
 rsaes_pkcs1_v1_5_encrypt rGen k m 
   | message_too_long = error "message too long"
   | otherwise        = (c, rGen')
  where
-  mLen = BS.length m
+  mLen = fromIntegral $ BS.length m
   kLen = public_size k
   -- Step 1
   message_too_long = mLen > (kLen - 11)
@@ -365,7 +367,7 @@ rsaes_pkcs1_v1_5_decrypt k c
   | signal_error = error "decryption error"
   | otherwise    = m
  where
-  mLen = BS.length c
+  mLen = fromIntegral $ BS.length c
   kLen = private_size k
   -- Step 1
   wrong_size = mLen /= kLen
@@ -419,7 +421,7 @@ rsassa_pkcs1_v1_5_verify hi k m s
  where
   kLen = public_size k
   -- Step 1
-  bad_size = BS.length s /= kLen
+  bad_size = BS.length s /= fromIntegral kLen
   -- Step 2
   s_i = os2ip s
   m_i = rsa_vp1 (public_n k) (public_e k) s_i
@@ -438,8 +440,8 @@ rsassa_pkcs1_v1_5_verify hi k m s
 -- 2^32 * hLen, where hLen is the passed length of the hash. 
 generate_MGF1 :: HashFunction -> MGF
 generate_MGF1 hash mgfSeed maskLen 
- | BS.length mgfSeed > ((2 ^ (32::Int64)) * hLen) = error "mask too long"
- | otherwise                                      = loop BS.empty 0
+ | BS.length mgfSeed > ((2 ^ (32::Int)) * hLen) = error "mask too long"
+ | otherwise                                    = loop BS.empty 0
  where
   hLen        = BS.length $ hash BS.empty
   end_counter = (maskLen `divCeil` hLen) - 1
@@ -507,7 +509,7 @@ sha256' = bytestringDigest . sha256
 
 -- "i2osp converts a nonnegative integer to an octet string of a specified
 -- length" -- RFC 3447
-i2osp :: Integral a => a -> Int64 -> ByteString
+i2osp :: Integral a => a -> Int -> ByteString
 i2osp x len | isTooLarge = error "RSA internal error: integer too large"
             | otherwise  = padding `BS.append` digits
  where
@@ -516,7 +518,7 @@ i2osp x len | isTooLarge = error "RSA internal error: integer too large"
   xAsInt     = fromIntegral x
   lenAsInt   = fromIntegral len
   --
-  padding = BS.replicate (len - BS.length digits) 0
+  padding = BS.replicate (fromIntegral len - BS.length digits) 0
   digits = BS.pack $ reverse $ digits256 x
   digits256 v 
     | v <= 255 = [fromIntegral v]
@@ -548,15 +550,15 @@ rsa_vp1 n e s
  | (s < 0) || (s >= n) = error "signature representative out of range"
  | otherwise           = modular_exponentiation s e n -- (s ^ e) `mod` n
  
-emsa_pkcs1_v1_5_encode :: HashInfo -> ByteString -> Int64 -> ByteString
+emsa_pkcs1_v1_5_encode :: HashInfo -> ByteString -> Int -> ByteString
 emsa_pkcs1_v1_5_encode (HashInfo hash_ident hash) m emLen 
-  | emLen < (tLen + 1) = error "intended encoded message length too short"
-  | otherwise          = em
+  | (fromIntegral emLen) < (tLen + 1) = error "intended encoded message length too short"
+  | otherwise                         = em
  where
   h = hash m
   t = hash_ident `BS.append` h
   tLen = BS.length t
-  ps = BS.replicate (emLen - tLen - 3) 0xFF
+  ps = BS.replicate (fromIntegral emLen - tLen - 3) 0xFF
   em = BS.concat [BS.singleton 0x00, BS.singleton 0x01, ps,
                   BS.singleton 0x00, t] 
 
@@ -576,21 +578,12 @@ chunkify len bstr
   | BS.length bstr <= len = [bstr]
   | otherwise             = (BS.take len bstr):(chunkify len $ BS.drop len bstr)
  
-#if !MIN_VERSION_random(1,0,1)
-instance Random Word8 where
-  randomR (a,b) g = let aI::Int = fromIntegral a 
-                        bI::Int = fromIntegral b
-                        (x, g') = randomR (aI, bI) g
-                    in (fromIntegral x, g')
-  random          = randomR (minBound, maxBound)
-#endif
-
-generate_random_bytestring :: RandomGen g => g -> Int64 -> (ByteString, g)
+generate_random_bytestring :: CryptoRandomGen g => g -> Int -> (ByteString, g)
 generate_random_bytestring g 0 = (BS.empty, g)
 generate_random_bytestring g x = (BS.cons' first rest, g'')
  where
   (rest, g')   = generate_random_bytestring g (x - 1)
-  (first, g'') = randomR (1,255) g' 
+  (first, g'') = throwLeft $ crandomR (1,255) g'
 
 -- Divide a by b, rounding towards positive infinity.
 divCeil :: Integral a => a -> a -> a
@@ -600,7 +593,7 @@ divCeil a b =
 
 -- Generate p and q. This is not necessarily the best way to do this, but the
 -- ANSI standard dealing with this cost money, and I was in a hurry.
-generate_pq :: RandomGen g => g -> Int64 -> (Integer, Integer, g)
+generate_pq :: CryptoRandomGen g => g -> Int -> (Integer, Integer, g)
 generate_pq g len 
   | len < 2   = error "length to short for generate_pq"
   | p == q    = generate_pq g'' len
@@ -610,7 +603,7 @@ generate_pq g len
   (baseQ, g'') = large_random_prime g' (len - (len `div` 2))
   (p, q)       = if baseP < baseQ then (baseQ, baseP) else (baseP, baseQ)
 
-large_random_prime :: RandomGen g => g -> Int64 -> (Integer, g)
+large_random_prime :: CryptoRandomGen g => g -> Int -> (Integer, g)
 large_random_prime g len = (prime, g''')
  where
   ([startH, startT], g') = random8s g 2
@@ -619,14 +612,14 @@ large_random_prime g len = (prime, g''')
   start                  = os2ip $ BS.pack start_ls
   (prime, g''')          = find_next_prime g'' start 
   
-random8s :: RandomGen g => g -> Int64 -> ([Word8], g)
+random8s :: CryptoRandomGen g => g -> Int -> ([Word8], g)
 random8s g 0 = ([], g)
 random8s g x = 
   let (rest, g') = random8s g (x - 1)
-      (next8, g'') = random g'
+      (next8, g'') = throwLeft (crandom g')
   in (next8:rest, g'')
 
-find_next_prime :: RandomGen g => g -> Integer -> (Integer, g)
+find_next_prime :: CryptoRandomGen g => g -> Integer -> (Integer, g)
 find_next_prime g n
   | even n             = error "Even number sent to find_next_prime"
   | n `mod` 65537 == 1 = find_next_prime g (n + 2)
@@ -635,7 +628,7 @@ find_next_prime g n
  where
   (got_a_prime, g') = is_probably_prime g n
 
-is_probably_prime :: RandomGen g => g -> Integer -> (Bool, g)
+is_probably_prime :: CryptoRandomGen g => g -> Integer -> (Bool, g)
 is_probably_prime !g !n 
   | any (\ x -> n `mod` x == 0) small_primes = (False, g)
   | otherwise                                = miller_rabin g n 20
@@ -658,12 +651,12 @@ is_probably_prime !g !n
                    877,  881,  883,  887,  907,  911,  919,  929,  937,  941,
                    947,  953,  967,  971,  977,  983,  991,  997, 1009, 1013  ]
 
-miller_rabin :: RandomGen g => g -> Integer -> Int -> (Bool, g)
+miller_rabin :: CryptoRandomGen g => g -> Integer -> Int -> (Bool, g)
 miller_rabin g _ 0             = (True, g)
 miller_rabin g n k | test a n  = (False, g')
                    | otherwise = miller_rabin g' n (k - 1)
  where
-  (a, g') = randomR (2, n - 2) g
+  (a, g') = throwLeft (crandomR (2, n - 2) g)
   base_b = tail $ reverse $ toBinary (n - 1) 
   -- 
   test a' n' = pow base_b a
